
#include "patternmatch.h"
#include <limits>
#include <cmath>
#include <iostream>
#include <omp.h>


int ComparePattern(const std::string &dataSet, const std::string &pattern, int offset)
{
  int result;
  const int patternLength = pattern.length();

  
  result = 1;
  for(int i = offset, patternIdx = 0; patternIdx < patternLength; i++, patternIdx++) {
    if(dataSet[i] != pattern[patternIdx]) {
      result = 0;
      break;
    }
  }

  return result;
}



int PatternCount(const std::string &dataSet, const std::string &pattern)
{
  int count = 0;
  const int loopLength = dataSet.length() - pattern.length();


  #pragma omp parallel for reduction(+:count)
  for (int i = 0; i <= loopLength; i++) {
    count += ComparePattern(dataSet, pattern, i);
  }
  
  return count;
}


tbb::concurrent_vector<int> PatternPositions(const std::string &dataSet, const std::string &pattern)
{
  tbb::concurrent_vector<int> positions;
  const int loopLength = dataSet.length() - pattern.length();


  #pragma omp parallel for
  for(int i = 0; i <= loopLength; i++) {
    
    if(ComparePattern(dataSet, pattern, i) > 0)
      positions.push_back(i);
  }

  return positions;
}


std::set<std::string> FrequentPatterns(const std::string &dataSet, int k)
{
  std::vector<int> counts((dataSet.length() - k) + 1);
  std::set<std::string> frequentPatterns;
  int t, maxCount;
  const int loopLength = dataSet.length() - k;

  
  #pragma omp declare reduction(maxcount : int : \
                                omp_out = omp_in > omp_out ? omp_in : omp_out \
				initializer(omp_priv=-1)

  maxCount = -1;
  frequentPatterns.empty();
  counts.empty();
  
  #pragma omp parallel for reduction(maxcount:maxCount)
  for(int i = 0; i <= loopLength; i++) {

    int t;
    t = PatternCount(dataSet, dataSet.substr(i, k));
    if(t > maxCount)
      maxCount = t;

    counts[i] = t;

  }


  for(int i = 0; i <= loopLength; i++) {
    if(counts[i] == maxCount)
      frequentPatterns.insert(dataSet.substr(i, k));
  }

  return frequentPatterns;

}


std::set<std::string> FindClumps(const std::string &dataSet, int L, int k, int t)
{
  std::set<std::string> kmerClumps;
  const int loopLength       = dataSet.length() - L;
  const int windowLoopLength = L - k;


  #pragma omp parallel for
  for(int startPos = 0; startPos <= loopLength; startPos++) {
    
    std::string window = dataSet.substr(startPos, L);
    for(int i = 0; i <= windowLoopLength; i++) {
      
      std::string pattern = window.substr(i, k);
      if(PatternCount(window, pattern) == t)
	kmerClumps.insert(pattern);
    }
  }

  return kmerClumps;

}


std::string ReverseCompliment(const std::string &dataSet)
{
  std::string rcomp;


  for(auto rit = dataSet.rbegin(); rit != dataSet.rend(); ++rit) {

    switch(*rit) {
      
    case 'a':
      rcomp.push_back('t');
      break;
    case 't':
      rcomp.push_back('a');
      break;
    case 'c':
      rcomp.push_back('g');
      break;
    case 'g':
      rcomp.push_back('c');
      break;

    }
  }

  return rcomp;
}


std::vector<int> MinSkews(const std::string &dataSet)
{
  int pos, skew, minSkew;
  std::vector<int> minSkews, skews;


  pos = 0;
  skew = 0;
  minSkew = std::numeric_limits<int>::max();
  for(auto it = dataSet.begin(); it != dataSet.end(); ++it) {

    switch(*it) {
      
    case 'c':
      skew--;
      break;
    case 'g':
      skew++;
      break;
    }

    if(skew < minSkew)
      minSkew = skew;

    skews.push_back(skew);

  }

  for(pos = 0; pos < skews.size(); pos++) {
    
    if(skews[pos] == minSkew) {
      minSkews.push_back(pos + 1);
    }
  }

  return minSkews;
}


int HammingDistance(const std::string &s1, const std::string &s2)
{
  int hDist;


  hDist = 0;
  for(int i = 0; i < s1.length(); i++) {
	  
    if(s1[i] != s2[i])
      hDist++;  
  }
  
  return hDist;
}


tbb::concurrent_vector<int> ApproximatePatternPositions(const std::string &dataSet, const std::string &pattern, int minHammingDistance)
{
  tbb::concurrent_vector<int> positions;
  const int patternLength = pattern.length();
  const int loopLength    = dataSet.length() - patternLength;


  #pragma omp parallel for
  for(int i = 0; i <= loopLength; i++) {

    int hDist = HammingDistance(dataSet.substr(i, patternLength), pattern);
    if(hDist <= minHammingDistance)
      positions.push_back(i);

  }

  return positions;
}


int ApproximatePatternCount(const std::string &dataSet, const std::string &pattern, int minHammingDistance)
{
  int count = 0;
  const int patternLength = pattern.length();
  const int loopLength = dataSet.length() - patternLength;


  #pragma omp parallel for reduction(+:count)
  for(int i = 0; i <= loopLength; i++) {

    int hDist = HammingDistance(dataSet.substr(i, patternLength), pattern);
    if(hDist <= minHammingDistance)
      count += 1;

  }

  return count;
}


int SymbolToNumber(char symbol)
{
  switch(symbol) {

  case 'a':
    return 0;
  case 'c':
    return 1;
  case 'g':
    return 2;
  case 't':
    return 3;
  }
}


std::string NumberToSymbol(int n)
{
  switch(n) {
  case 0:
    return "a";
  case 1:
    return "c";
  case 2:
    return "g";
  case 3:
    return "t";
  }
}


int PatternToNumber(const std::string &pattern)
{
  char lastSymbol;
  std::string prefix;


  if(pattern.length() == 0)
    return 0;

  lastSymbol = pattern[pattern.length() - 1];

  if(pattern.length() == 1)
    prefix = "";
  else
    prefix = pattern.substr(0, pattern.length() - 1);

  return(4 * PatternToNumber(prefix) + SymbolToNumber(lastSymbol));
}


std::string NumberToPattern(int i, int k)
{
  int q, r;
  std::string prefixPattern;
  std::string symbol;


  if(k == 1) {
    return NumberToSymbol(i);
  }

  q = i / 4;
  r = i % 4;
  symbol = NumberToSymbol(r);
  prefixPattern = NumberToPattern(q, k - 1);
  return(prefixPattern + symbol);
}


std::string Suffix(const std::string &pattern)
{
  std::string suffix;

  if(pattern.length() == 1) {
    suffix = "";
  }
  else {
    suffix = pattern.substr(1);
  }

  return suffix;
}


std::set<std::string> Neighbors(const std::string &pattern, int minHammingDistance)
{
  std::set<std::string> result;
  std::set<std::string> suffixNeighbors;
  std::set<std::string> neighborhood;
  

  // Base case 
  if(minHammingDistance == 0) {
    result.insert(pattern);
    return result;
  }

  if(pattern.length() == 1) {
    result.insert({ "a", "c", "g", "t" });
    return result;
  }

  neighborhood.empty();
  suffixNeighbors = Neighbors(Suffix(pattern), minHammingDistance);
  for(auto s : suffixNeighbors) {
    if(HammingDistance(Suffix(pattern), s) < minHammingDistance) {
      neighborhood.insert("a" + s);
      neighborhood.insert("c" + s);
      neighborhood.insert("g" + s);
      neighborhood.insert("t" + s);
    }
    else {
      neighborhood.insert(pattern[0] + s);
    }
  }

  return neighborhood;
}


std::set<std::string> FrequentPatternsWithMismatches(const std::string &dataSet, int k, int minHammingDistance)
{
  const int arrSize = (int)pow(4.0, k);
  std::set<std::string> frequentPatterns;
  std::vector<int> close(arrSize);
  std::vector<int> frequencyArray(arrSize);
  int maxFreqCount;
  const int loopLength = dataSet.length() - k;


  #pragma omp declare reduction(maxcount : int : \
                                omp_out = omp_in > omp_out ? omp_in : omp_out \
				initializer(omp_priv=-1)

  frequentPatterns.empty();
  for(int i = 0; i < arrSize; i++) {
    close.push_back(0);
    frequencyArray.push_back(0);
  }

  #pragma omp parallel for
  for(int i = 0; i <= loopLength; i++) {

    std::set<std::string> neighborhood = Neighbors(dataSet.substr(i, k), minHammingDistance);
    for(auto pattern : neighborhood) {
      close[PatternToNumber(pattern)] = 1;
    }
  }

  #pragma omp parallel for
  for(int i = 0; i < arrSize; i++) {
    if(close[i] == 1) {
      frequencyArray[i] = ApproximatePatternCount(dataSet, NumberToPattern(i, k), minHammingDistance);
    }
  }


  #pragma omp parallel for reduction(maxcount:maxFreqCount)
  for(int i = 0; i < arrSize; i++) {
    if(frequencyArray[i] > maxFreqCount)
      maxFreqCount = frequencyArray[i];
  }


  if(maxFreqCount > 0) {
    for(int i = 0; i < arrSize; i++) {
      if(frequencyArray[i] == maxFreqCount) {
	frequentPatterns.insert(NumberToPattern(i, k));
      }
    }
  }

  return frequentPatterns;
}


std::set<std::string> FrequentPatternsRCWithMismatches(const std::string &dataSet, int k, int minHammingDistance)
{
  const int arrSize = (int)pow(4.0, k);
  std::set<std::string> frequentPatterns;
  std::vector<int> close(arrSize);
  std::vector<int> frequencyArray(arrSize);
  int maxFreqCount;
  const int loopLength = dataSet.length() - k;


  #pragma omp declare reduction(maxcount : int : \
                                omp_out = omp_in > omp_out ? omp_in : omp_out \
				initializer(omp_priv=-1)

  frequentPatterns.empty();
  for(int i = 0; i < arrSize; i++) {
    close.push_back(0);
    frequencyArray.push_back(0);
  }

  #pragma omp parallel for
  for(int i = 0; i <= loopLength; i++) {

    std::set<std::string> neighborhood = Neighbors(dataSet.substr(i, k), minHammingDistance);
    for(auto pattern : neighborhood) {
      close[PatternToNumber(pattern)] = 1;
    }
  }

  #pragma omp parallel for
  for(int i = 0; i < arrSize; i++) {
    if(close[i] == 1) {
      
      int pCount, rcompCount;
      
      pCount     = ApproximatePatternCount(dataSet, NumberToPattern(i, k), minHammingDistance);
      rcompCount = ApproximatePatternCount(dataSet, ReverseCompliment(NumberToPattern(i, k)), minHammingDistance);
      frequencyArray[i] = pCount + rcompCount;
    }
  }


  #pragma omp parallel for reduction(maxcount:maxFreqCount)
  for(int i = 0; i < arrSize; i++) {
    if(frequencyArray[i] > maxFreqCount)
      maxFreqCount = frequencyArray[i];
  }

  if(maxFreqCount > 0) {
    for(int i = 0; i < arrSize; i++) {
      if(frequencyArray[i] == maxFreqCount) {
	frequentPatterns.insert(NumberToPattern(i, k));
	frequentPatterns.insert(ReverseCompliment(NumberToPattern(i, k)));
      }
    }
  }

  return frequentPatterns;
}
